import { EMPTY, firstValueFrom, from, interval, Observable, of, Subject } from 'rxjs';
import { catchError, debounce, finalize, first, map, mergeWith, share, switchMap, take } from 'rxjs/operators';
import { RpcServerError } from './constants';
import { BufferSubject } from '../../../util/BufferSubject';
import { RpcError, RpcValidationError } from './error';
export class RpcApiCaller {
    constructor({ api, preCallBufferSize = 10, maxActiveCalls = 50, }) {
        this._calls = 0;
        this.api = api;
        this.preCallBufferSize = preCallBufferSize;
        this.maxActiveCalls = maxActiveCalls;
    }
    get activeCalls() {
        return this._calls;
    }
    exists(name) {
        return this.api.hasOwnProperty(name);
    }
    get(name) {
        if (!this.exists(name))
            throw new RpcError(RpcServerError.NoMethodSpecified);
        return this.api[name];
    }
    async call(name, request, ctx) {
        if (this._calls >= this.maxActiveCalls)
            throw new RpcError(RpcServerError.TooManyActiveCalls);
        const method = this.get(name);
        if (method.validate) {
            try {
                method.validate(request);
            }
            catch (error) {
                throw new RpcValidationError(error);
            }
        }
        this._calls++;
        return (method.onPreCall ? method.onPreCall(ctx, request) : Promise.resolve())
            .then(() => !method.isStreaming
            ? method.call(ctx, request)
            : firstValueFrom(method.call$(ctx, of(request))))
            .finally(() => {
            this._calls--;
        });
    }
    createCall(name, ctx) {
        const req$ = new Subject();
        const reqUnsubscribe$ = new Subject();
        try {
            if (this._calls >= this.maxActiveCalls)
                throw new RpcError(RpcServerError.TooManyActiveCalls);
            const method = this.get(name);
            if (!method.isStreaming) {
                const response$ = from((async () => {
                    const request = await firstValueFrom(req$.pipe(first()));
                    const response = await this.call(name, request, ctx);
                    return response;
                })());
                const res$ = new Subject();
                response$.subscribe(res$);
                return { req$, reqUnsubscribe$, res$ };
            }
            const methodStreaming = method;
            const requestValidated$ = req$
                .pipe(map(request => {
                try {
                    if (methodStreaming.validate)
                        methodStreaming.validate(request);
                    return request;
                }
                catch (error) {
                    throw new RpcValidationError(error);
                }
            }));
            const bufferSize = methodStreaming.preCallBufferSize || this.preCallBufferSize;
            const requestBuffered$ = new BufferSubject(bufferSize);
            requestBuffered$.subscribe({ error: () => { } });
            const error$ = new Subject();
            requestBuffered$.subscribe({
                error: error => { error$.error(error); },
            });
            requestValidated$.subscribe(requestBuffered$);
            const result$ = requestBuffered$
                .pipe(take(1), switchMap(request => {
                return methodStreaming.onPreCall
                    ? from(methodStreaming.onPreCall(ctx, request))
                    : from([0]);
            }), switchMap(() => {
                Promise.resolve().then(() => {
                    requestBuffered$.flush();
                });
                return methodStreaming.call$(ctx, requestBuffered$)
                    .pipe(finalize(() => {
                    error$.complete();
                }));
            }), share(), mergeWith(error$));
            this._calls++;
            const resultWithActiveCallTracking$ = result$.pipe(finalize(() => {
                error$.complete();
                this._calls--;
            }));
            const res$ = new Observable(observer => {
                const subscription = resultWithActiveCallTracking$.subscribe(observer);
                const timeout = methodStreaming.timeout ?? 15_000;
                const timeoutSubscription = requestBuffered$
                    .pipe(catchError(() => EMPTY), mergeWith(result$.pipe(catchError(() => EMPTY))), debounce(() => interval(timeout)))
                    .subscribe(() => {
                    const error = new RpcError(RpcServerError.Timeout);
                    error$.error(error);
                });
                return () => {
                    timeoutSubscription.unsubscribe();
                    subscription.unsubscribe();
                };
            });
            return { req$, reqUnsubscribe$, res$ };
        }
        catch (error) {
            req$.error(error);
            const res$ = new Subject();
            res$.error(error);
            return { req$, reqUnsubscribe$, res$ };
        }
    }
    call$(name, request$, ctx) {
        const call = this.createCall(name, ctx);
        request$.subscribe(call.req$);
        return call.res$;
    }
}
